import numpy as np
import multiprocess as mp

import datagen_modules
import DML_modules
import data_generator_2
import data_generator_1
import run_Experiment

import sys 

# Generate random permutation of features of

def generateRandomPerm(X):
    # Making a random seed
    np.random.seed()
    nlist = np.sort(np.asarray(X.columns))
    nlist_random = np.random.permutation(nlist)
    Slist = []
    for idx in range(len(nlist_random)+1):
        Slist.append( nlist_random[:idx] )
    return Slist

def stopCriterion(accumDict, curDict, stop_value):
    for key in accumDict.keys():
        if np.abs( accumDict[key] - curDict[key] ) > stop_value:
            return False
    return True

# Combine individual permutation results
def combineResult(results):
    result_dict = dict()
    for key in results[0].keys():
        result_dict[key] = 0

    for idx in range(len(results)):
        for key in results[idx].keys():
            result_dict[key] =  ((idx * result_dict[key]) + results[idx][key]) / (idx+1)
    return result_dict

##########################################################################################
##########################################################################################
# DML
##########################################################################################
##########################################################################################

##########################################################################################
# Case 1
##########################################################################################
# Update of DMLDict(S) for individual permutation
def Shapley_update_perm(X_list, y_list, allCondProb_list, x, y, Slist, topoSort, noiseTF, DMLDict):
    '''
    Output:
        E[Y | do(v_{S1})] - E[Y | do(v_{S0})]
        Updated DMLDict
    '''
    # Dictionary Dict(Var) for individual permutation
    shapley_dict_perm = dict()
    # For each S generated for the individual permutation
    for idx in range(1,len(X_list[0].columns)+1):
        # pre_{\pi}(Vi)
        S0 = Slist[idx-1]
        # Vi and pre_{\pi}(Vi)
        S1 = Slist[idx]
        # Vi
        updated_variable = np.setdiff1d(S1,S0)[0]

        # If S0 = emptyset,
        if len(S0) == 0:
            # E[Y]
            result_S0 = (np.mean(y_list[0]) + np.mean(y_list[1]))/2
        else:
            # Update the DMLDict for S0
            DMLDict = DML_modules.computeDML_S(X_list, y_list, allCondProb_list, x, S0, topoSort, DMLDict, noiseTF)
            # Have E[Y | do(v_{S0})]
            S0_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S0)
            result_S0 = DMLDict[DML_modules.keygen(S0_sorted)]

        # If S1 = [n], give y
        if len(S1) == len(X_list[0].columns):
            result_S1 = y
        else:
            # Update the DMLDict for S1
            DMLDict = DML_modules.computeDML_S(X_list, y_list, allCondProb_list, x, S1, topoSort, DMLDict, noiseTF)
            # Have E[Y | do(v_{S1})]
            S1_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S1)
            result_S1 = DMLDict[DML_modules.keygen(S1_sorted)]

        # Compute E[Y | do(v_{S1})] - E[Y | do(v_{S0})]
        shapley_dict_perm[updated_variable] = result_S1 - result_S0

    return shapley_dict_perm, DMLDict

def computeShapley_DML(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter, noiseTF):
    '''
    Output
        shapley_dict: Final Shapley Value
    '''
    # Initialize the DML_Dict
    DML_dict = dict()
    # Final Shapley Value
    shapley_dict = dict()

    # allCondModel trained using Train and Test dataset
    allCondModel_train = DML_modules.train_condProb(X_train, topoSort)
    allCondModel_test = DML_modules.train_condProb(X_test, topoSort)

    # Evaluated {P(Vi | pre(Vi)) } trained using Train and Test dataset
    allCondProb_eval_test = DML_modules.estimate_condProb(X_test, allCondModel_train, topoSort, noiseTF)
    allCondProb_eval_train = DML_modules.estimate_condProb(X_train, allCondModel_test, topoSort, noiseTF)
    X_list = [X_train, X_test]
    y_list = [y_train, y_test]
    allCondProb_list = [allCondProb_eval_train, allCondProb_eval_test]

    for key in X_train.columns:
        shapley_dict[key] = 0

    # Repeat tje following:
    for idx in range(numiter):
        # Generate S
        Slist = generateRandomPerm(X_train)
        # Generate an individual Shapley S and DML Dict.
        shapley_dict_S, DML_dict = Shapley_update_perm(X_list, y_list, allCondProb_list, x, y, Slist, topoSort, noiseTF, DML_dict)

        # Moving Average Update for the Shapley.
        for key in X_train.columns:
            shapley_dict[key] = ((idx * shapley_dict[key]) + shapley_dict_S[key]) / (idx+1)

    return shapley_dict

def multi_run_wrapper_DML(args):
    return computeShapley_DML(*args)

def Run_DML(n,seednum,numiter_per_sim,num_sim,targetidx,num_process,noiseTF):
    '''
    num_sim = number of parallelized worker
    numiter_per_sim = For each parallelized worker, the number of random permutation
    That is, the total number is num_sim *numiter_per_sim.
    '''
    X, topoSort, dictName, parentDict = data_generator_1.dataGen(n, seednum)
    X_train, y_train, X_test, y_test = datagen_modules.dataSplit(X, seednum)
    x = X.iloc[[targetidx]]
    y = x[x.columns[-1]].values[0]
    x = x[x.columns[:-1]]

    numiter = numiter_per_sim * num_sim

    resultShapley = computeShapley_DML(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter, noiseTF)

    # myparam = (X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter_per_sim)

    # pool = mp.Pool(processes=num_process)
    # results = pool.map(multi_run_wrapper_DML, [myparam] * num_sim)
    # resultShapley = combineResult(results)
    # for idx in range(len(X_train.columns)):
    #     print(dictName[idx], ":", np.round(resultShapley[idx], 4))
    return resultShapley


##########################################################################################
# Case 2
##########################################################################################

# Update of DMLDict(S) for individual permutation
def Shapley_update_perm_case2(X_list, y_list, model_Y_list, x, y, Slist, topoSort, noiseTF, DMLDict):
    '''
    Output:
        E[Y | do(v_{S1})] - E[Y | do(v_{S0})]
        Updated DMLDict
    '''
    # Dictionary Dict(Var) for individual permutation
    shapley_dict_perm = dict()
    # For each S generated for the individual permutation
    for idx in range(1,len(X_list[0].columns)+1):
        # pre_{\pi}(Vi)
        S0 = Slist[idx-1]
        # Vi and pre_{\pi}(Vi)
        S1 = Slist[idx]
        # Vi
        updated_variable = np.setdiff1d(S1,S0)[0]

        # If S0 = emptyset,
        if len(S0) == 0:
            # E[Y]
            result_S0 = (np.mean(y_list[0]) + np.mean(y_list[1]))/2
        else:
            # Update the DMLDict for S0
            DMLDict = DML_modules.computeDML_S_Case2(X_list, y_list, model_Y_list, x, S0, topoSort, DMLDict, noiseTF)
            # Have E[Y | do(v_{S0})]
            S0_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S0)
            result_S0 = DMLDict[DML_modules.keygen(S0_sorted)]

        # If S1 = [n], give y
        if len(S1) == len(X_list[0].columns):
            result_S1 = y
        else:
            # Update the DMLDict for S1
            DMLDict = DML_modules.computeDML_S_Case2(X_list, y_list, model_Y_list, x, S1, topoSort, DMLDict, noiseTF)
            # Have E[Y | do(v_{S1})]
            S1_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S1)
            result_S1 = DMLDict[DML_modules.keygen(S1_sorted)]

        # Compute E[Y | do(v_{S1})] - E[Y | do(v_{S0})]
        shapley_dict_perm[updated_variable] = result_S1 - result_S0

    return shapley_dict_perm, DMLDict

# Compute Shapley Value
def computeShapley_DML_case2(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter, noiseTF):
    '''
    Output
        shapley_dict: Final Shapley Value
    '''
    # Initialize the DML_Dict
    DML_dict = dict()

    # Final Shapley Value
    shapley_dict = dict()

    X_list = [X_train, X_test]
    y_list = [y_train, y_test]

    model_Y_train = DML_modules.train_Y(X_train, y_train)
    model_Y_test = DML_modules.train_Y(X_test, y_test)
    model_Y_list = [model_Y_train, model_Y_test]

    for key in X_train.columns:
        shapley_dict[key] = 0

    # Repeat tje following:
    for idx in range(numiter):
        # Generate S
        Slist = generateRandomPerm(X_train)
        # Generate an individual Shapley S and DML Dict.
        shapley_dict_S, DML_dict = Shapley_update_perm_case2(X_list, y_list, model_Y_list, x, y, Slist, topoSort, noiseTF, DML_dict)

        # Moving Average Update for the Shapley.
        for key in X_train.columns:
            shapley_dict[key] = ((idx * shapley_dict[key]) + shapley_dict_S[key]) / (idx + 1)

    return shapley_dict


def multi_run_wrapper_DML_case2(args):
    return computeShapley_DML_case2(*args)

def Run_DML_case2(n,seednum,numiter_per_sim,num_sim,targetidx,num_process, noiseTF):
    '''
    num_sim = number of parallelized worker
    numiter_per_sim = For each parallelized worker, the number of random permutation
    That is, the total number is num_sim *numiter_per_sim.
    '''
    X, topoSort, dictName, parentDict = data_generator_2.dataGen(n, seednum)
    X_train, y_train, X_test, y_test = datagen_modules.dataSplit(X, seednum)
    x = X.iloc[[targetidx]]
    y = x[x.columns[-1]].values[0]
    x = x[x.columns[:-1]]
    numiter = numiter_per_sim * num_sim 

    resultShapley = computeShapley_DML_case2(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter, noiseTF)

    # myparam = (X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter_per_sim)

    # pool = mp.Pool(processes=num_process)
    # results = pool.map(multi_run_wrapper_DML_case2, [myparam] * num_sim)
    # resultShapley = combineResult(results)
    # for idx in range(len(X_train.columns)):
    #     print(dictName[idx], ":", np.round(resultShapley[idx], 4))
    return resultShapley



##########################################################################################
##########################################################################################
# IPW
##########################################################################################
##########################################################################################

##########################################################################################
# Case 1
##########################################################################################
# Update of IPWDict(S) for individual permutation
def Shapley_update_perm_IPW(X_list, y_list, allCondProb_list, x, y, Slist, topoSort, noiseTF, IPWDict):
    shapley_dict_perm = dict()
    for idx in range(1,len(X_list[0].columns)+1):
        S0 = Slist[idx-1]
        S1 = Slist[idx]
        updated_variable = np.setdiff1d(S1,S0)[0]
        if len(S0) == 0:
            result_S0 = (np.mean(y_list[0]) + np.mean(y_list[1]))/2
        else:
            IPWDict = DML_modules.computeIPW_S(X_list, y_list, allCondProb_list, x, S0, topoSort, IPWDict)
            S0_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S0)
            result_S0 = IPWDict[DML_modules.keygen(S0_sorted)]
            # result_S0 = DML_modules.computeDML_S(X_list,y_list, allCondProb_list, x, S0, topoSort, DMLDict)
        # print("S0:", S0)
        if len(S1) == len(X_list[0].columns):
            result_S1 = y
        else:
            # print("3", DMLDict)
            IPWDict = DML_modules.computeIPW_S(X_list, y_list, allCondProb_list, x, S1, topoSort, IPWDict)
            # print("4", DMLDict)
            S1_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S1)
            result_S1 = IPWDict[DML_modules.keygen(S1_sorted)]
            # result_S1 = DML_modules.computeDML(X_train, y_train, X_test, y_test, x, S1, parentDict, topoSort)
        # print("S1:", S1)
        shapley_dict_perm[updated_variable] = result_S1 - result_S0

    return shapley_dict_perm, IPWDict

def computeShapley_IPW(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter, noiseTF):
    IPW_dict = dict()
    shapley_dict = dict()

    # allCondModel trained using Train and Test dataset
    allCondModel_train = DML_modules.train_condProb(X_train, topoSort)
    allCondModel_test = DML_modules.train_condProb(X_test, topoSort)

    # Evaluated {P(Vi | pre(Vi)) } trained using Train and Test dataset
    allCondProb_eval_test = DML_modules.estimate_condProb(X_test, allCondModel_train, topoSort, noiseTF)
    allCondProb_eval_train = DML_modules.estimate_condProb(X_train, allCondModel_test, topoSort, noiseTF)
    X_list = [X_train, X_test]
    y_list = [y_train, y_test]
    allCondProb_list = [allCondProb_eval_train, allCondProb_eval_test]

    # prevDict = dict()
    for key in X_train.columns:
        shapley_dict[key] = 0
        # prevDict[key] = 0

    for idx in range(numiter):
        # print(idx)
        Slist = generateRandomPerm(X_train)
        shapley_dict_S, IPW_dict = Shapley_update_perm_IPW(X_list, y_list, allCondProb_list, x, y, Slist, topoSort, noiseTF, IPW_dict)

        for key in X_train.columns:
            shapley_dict[key] = ((idx * shapley_dict[key]) + shapley_dict_S[key]) / (idx + 1)

    return shapley_dict

def multi_run_wrapper_IPW(args):
    return computeShapley_IPW(*args)

def Run_IPW(n,seednum,numiter_per_sim,num_sim,targetidx,num_process, noiseTF):
    X, topoSort, dictName, parentDict = data_generator_1.dataGen(n, seednum)
    X_train, y_train, X_test, y_test = datagen_modules.dataSplit(X, seednum)
    x = X.iloc[[targetidx]]
    y = x[x.columns[-1]].values[0]
    x = x[x.columns[:-1]]

    myparam = (X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter_per_sim)
    numiter = numiter_per_sim * num_sim
    resultShapley = computeShapley_IPW(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter, noiseTF)

    # pool = mp.Pool(processes=num_process)
    # results = pool.map(multi_run_wrapper_IPW, [myparam] * num_sim)
    # resultShapley = combineResult(results)
    # for idx in range(len(X_train.columns)):
    #     print(dictName[idx], ":", np.round(resultShapley[idx], 4))
    return resultShapley

##########################################################################################
# Case 2
##########################################################################################

# Update of IPWDict(S) for individual permutation
def Shapley_update_perm_IPW_case2(X_list, y_list, x, y, Slist, topoSort, noiseTF, IPWDict):
    shapley_dict_perm = dict()
    for idx in range(1,len(X_list[0].columns)+1):
        S0 = Slist[idx-1]
        S1 = Slist[idx]
        updated_variable = np.setdiff1d(S1,S0)[0]
        if len(S0) == 0:
            result_S0 = (np.mean(y_list[0]) + np.mean(y_list[1]))/2
        else:
            IPWDict = DML_modules.computeIPW_S_case2(X_list, y_list,  x, S0, topoSort, IPWDict, noiseTF)
            S0_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S0)
            result_S0 = IPWDict[DML_modules.keygen(S0_sorted)]
            # result_S0 = DML_modules.computeDML_S(X_list,y_list, allCondProb_list, x, S0, topoSort, DMLDict)
        # print("S0:", S0)
        if len(S1) == len(X_list[0].columns):
            result_S1 = y
        else:
            # print("3", DMLDict)
            IPWDict = DML_modules.computeIPW_S_case2(X_list, y_list,  x, S1, topoSort, IPWDict, noiseTF)
            # print("4", DMLDict)
            S1_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S1)
            result_S1 = IPWDict[DML_modules.keygen(S1_sorted)]
            # result_S1 = DML_modules.computeDML(X_train, y_train, X_test, y_test, x, S1, parentDict, topoSort)
        # print("S1:", S1)
        shapley_dict_perm[updated_variable] = result_S1 - result_S0

    return shapley_dict_perm, IPWDict

def computeShapley_IPW_case2(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter, noiseTF):
    IPW_dict = dict()
    shapley_dict = dict()

    X_list = [X_train, X_test]
    y_list = [y_train, y_test]

    # prevDict = dict()
    for key in X_train.columns:
        shapley_dict[key] = 0
        # prevDict[key] = 0

    for idx in range(numiter):
        # print(idx)
        Slist = generateRandomPerm(X_train)
        shapley_dict_S, IPW_dict = Shapley_update_perm_IPW_case2(X_list, y_list, x, y, Slist, topoSort, noiseTF, IPW_dict)

        for key in X_train.columns:
            shapley_dict[key] = ((idx * shapley_dict[key]) + shapley_dict_S[key]) / (idx + 1)

    return shapley_dict

def multi_run_wrapper_IPW_case2(args):
    return computeShapley_IPW_case2(*args)

def Run_IPW_case2(n,seednum,numiter_per_sim,num_sim,targetidx,num_process, noiseTF):
    X, topoSort, dictName, parentDict = data_generator_2.dataGen(n, seednum)
    X_train, y_train, X_test, y_test = datagen_modules.dataSplit(X, seednum)
    x = X.iloc[[targetidx]]
    y = x[x.columns[-1]].values[0]
    x = x[x.columns[:-1]]

    myparam = (X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter_per_sim)
    numiter = numiter_per_sim * num_sim
    resultShapley = computeShapley_IPW_case2(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter, noiseTF)

    # myparam = (X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter_per_sim)

    # pool = mp.Pool(processes=num_process)
    # results = pool.map(multi_run_wrapper_IPW_case2, [myparam] * num_sim)
    # resultShapley = combineResult(results)
    # for idx in range(len(X_train.columns)):
    #     print(dictName[idx], ":", np.round(resultShapley[idx], 4))
    return resultShapley

##########################################################################################
##########################################################################################
# PI
##########################################################################################
##########################################################################################

##########################################################################################
# Case 1
##########################################################################################

# Update of PIDict(S) for individual permutation
def Shapley_update_perm_PI(X_list, y_list, allCondProb_list, x, y, Slist, topoSort, noiseTF, PIDict):
    shapley_dict_perm = dict()
    for idx in range(1,len(X_list[0].columns)+1):
        S0 = Slist[idx-1]
        S1 = Slist[idx]
        updated_variable = np.setdiff1d(S1,S0)[0]
        if len(S0) == 0:
            result_S0 = (np.mean(y_list[0]) + np.mean(y_list[1]))/2
        else:
            PIDict = DML_modules.computePI_S(X_list,y_list, x, S0, topoSort, PIDict, noiseTF)
            S0_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S0)
            result_S0 = PIDict[DML_modules.keygen(S0_sorted)]
            # result_S0 = DML_modules.computeDML_S(X_list,y_list, allCondProb_list, x, S0, topoSort, DMLDict)
        # print("S0:", S0)
        if len(S1) == len(X_list[0].columns):
            result_S1 = y
        else:
            # print("3", DMLDict)
            PIDict = DML_modules.computePI_S(X_list,y_list, x, S1, topoSort, PIDict, noiseTF)
            # print("4", DMLDict)
            S1_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S1)
            result_S1 = PIDict[DML_modules.keygen(S1_sorted)]
            # result_S1 = DML_modules.computeDML(X_train, y_train, X_test, y_test, x, S1, parentDict, topoSort)
        # print("S1:", S1)
        shapley_dict_perm[updated_variable] = result_S1 - result_S0

    return shapley_dict_perm, PIDict

def computeShapley_PI(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter, noiseTF):
    PI_Dict = dict()
    shapley_dict = dict()

    # allCondModel trained using Train and Test dataset
    allCondModel_train = DML_modules.train_condProb(X_train, topoSort)
    allCondModel_test = DML_modules.train_condProb(X_test, topoSort)

    # Evaluated {P(Vi | pre(Vi)) } trained using Train and Test dataset
    allCondProb_eval_test = DML_modules.estimate_condProb(X_test, allCondModel_train, topoSort, noiseTF)
    allCondProb_eval_train = DML_modules.estimate_condProb(X_train, allCondModel_test, topoSort, noiseTF)
    X_list = [X_train, X_test]
    y_list = [y_train, y_test]
    allCondProb_list = [allCondProb_eval_train, allCondProb_eval_test]

    # prevDict = dict()
    for key in X_train.columns:
        shapley_dict[key] = 0
        # prevDict[key] = 0

    for idx in range(numiter):
        # print(idx)
        Slist = generateRandomPerm(X_train)
        shapley_dict_S, PI_Dict = Shapley_update_perm_PI(X_list, y_list, allCondProb_list, x, y, Slist, topoSort, noiseTF, PI_Dict)

        for key in X_train.columns:
            shapley_dict[key] = ((idx * shapley_dict[key]) + shapley_dict_S[key]) / (idx + 1)

    return shapley_dict

def multi_run_wrapper_PI(args):
    return computeShapley_PI(*args)

def Run_PI(n,seednum,numiter_per_sim,num_sim,targetidx,num_process, noiseTF):
    X, topoSort, dictName, parentDict = data_generator_1.dataGen(n, seednum)
    X_train, y_train, X_test, y_test = datagen_modules.dataSplit(X, seednum)
    x = X.iloc[[targetidx]]
    y = x[x.columns[-1]].values[0]
    x = x[x.columns[:-1]]

    myparam = (X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter_per_sim)
    numiter = numiter_per_sim * num_sim
    resultShapley = computeShapley_PI(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter, noiseTF)

    # myparam = (X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter_per_sim)

    # pool = mp.Pool(processes=num_process)
    # results = pool.map(multi_run_wrapper_PI, [myparam] * num_sim)
    # resultShapley = combineResult(results)
    # for idx in range(len(X_train.columns)):
    #     print(dictName[idx], ":", np.round(resultShapley[idx], 4))
    return resultShapley


##########################################################################################
# Case 2
##########################################################################################

# Update of PIDict(S) for individual permutation
def Shapley_update_perm_PI_case2(X_list, y_list, model_Y_list, x, y, Slist, topoSort, noiseTF, PIDict):
    shapley_dict_perm = dict()
    for idx in range(1,len(X_list[0].columns)+1):
        S0 = Slist[idx-1]
        S1 = Slist[idx]
        updated_variable = np.setdiff1d(S1,S0)[0]
        if len(S0) == 0:
            result_S0 = (np.mean(y_list[0]) + np.mean(y_list[1]))/2
        else:
            PIDict = DML_modules.computePI_S_case2(X_list,y_list, model_Y_list, x, S0, topoSort, PIDict, noiseTF)
            S0_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S0)
            result_S0 = PIDict[DML_modules.keygen(S0_sorted)]
            # result_S0 = DML_modules.computeDML_S(X_list,y_list, allCondProb_list, x, S0, topoSort, DMLDict)
        # print("S0:", S0)
        if len(S1) == len(X_list[0].columns):
            result_S1 = y
        else:
            # print("3", DMLDict)
            PIDict =DML_modules.computePI_S_case2(X_list,y_list, model_Y_list, x, S1, topoSort, PIDict, noiseTF)
            # print("4", DMLDict)
            S1_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S1)
            result_S1 = PIDict[DML_modules.keygen(S1_sorted)]
            # result_S1 = DML_modules.computeDML(X_train, y_train, X_test, y_test, x, S1, parentDict, topoSort)
        # print("S1:", S1)
        shapley_dict_perm[updated_variable] = result_S1 - result_S0

    return shapley_dict_perm, PIDict

def computeShapley_PI_case2(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter, noiseTF):
    PI_Dict = dict()
    shapley_dict = dict()

    X_list = [X_train, X_test]
    y_list = [y_train, y_test]

    # Model_Y
    model_Y_train = DML_modules.train_Y(X_train, y_train)
    model_Y_test = DML_modules.train_Y(X_test, y_test)
    model_Y_list = [model_Y_train, model_Y_test]

    # prevDict = dict()
    for key in X_train.columns:
        shapley_dict[key] = 0
        # prevDict[key] = 0

    for idx in range(numiter):
        # print(idx)
        Slist = generateRandomPerm(X_train)
        shapley_dict_S, PI_Dict = Shapley_update_perm_PI_case2(X_list, y_list, model_Y_list, x, y, Slist, topoSort, noiseTF, PI_Dict)

        for key in X_train.columns:
            shapley_dict[key] = ((idx * shapley_dict[key]) + shapley_dict_S[key]) / (idx + 1)

    return shapley_dict

def multi_run_wrapper_PI_case2(args):
    return computeShapley_PI_case2(*args)

def Run_PI_case2(n,seednum,numiter_per_sim,num_sim,targetidx,num_process,noiseTF):
    X, topoSort, dictName, parentDict = data_generator_2.dataGen(n, seednum)
    X_train, y_train, X_test, y_test = datagen_modules.dataSplit(X, seednum)
    x = X.iloc[[targetidx]]
    y = x[x.columns[-1]].values[0]
    x = x[x.columns[:-1]]

    myparam = (X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter_per_sim)
    numiter = numiter_per_sim * num_sim
    resultShapley = computeShapley_PI_case2(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter, noiseTF)

    # myparam = (X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter_per_sim)
    # numiter = numiter_per_sim * num_sim
    # resultShapley = computeShapley_IPW(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter)

    # myparam = (X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter_per_sim)

    # pool = mp.Pool(processes=num_process)
    # results = pool.map(multi_run_wrapper_PI, [myparam] * num_sim)
    # resultShapley = combineResult(results)
    # for idx in range(len(X_train.columns)):
    #     print(dictName[idx], ":", np.round(resultShapley[idx], 4))
    return resultShapley

##########################################################################################
##########################################################################################
# Sample
##########################################################################################
##########################################################################################

# Update of PIDict(S) for individual permutation
def Shapley_update_perm_Sample(X_list, y_list, allCondModel_list, model_Y_list, x, y, Slist, topoSort, parentDict, noiseTF, SampleDict):
    shapley_dict_perm = dict()
    for idx in range(1,len(X_list[0].columns)+1):
        S0 = Slist[idx-1]
        S1 = Slist[idx]
        updated_variable = np.setdiff1d(S1,S0)[0]
        if len(S0) == 0:
            result_S0 = (np.mean(y_list[0]) + np.mean(y_list[1]))/2
        else:
            SampleDict = DML_modules.computeSampling_S(X_list,y_list, allCondModel_list, model_Y_list, x, S0, topoSort, parentDict, SampleDict, noiseTF)
            S0_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S0)
            result_S0 = SampleDict[DML_modules.keygen(S0_sorted)]
            # result_S0 = DML_modules.computeDML_S(X_list,y_list, allCondProb_list, x, S0, topoSort, DMLDict)
        # print("S0:", S0)
        if len(S1) == len(X_list[0].columns):
            result_S1 = y
        else:
            # print("3", DMLDict)
            SampleDict = DML_modules.computeSampling_S(X_list,y_list, allCondModel_list, model_Y_list, x, S1, topoSort, parentDict, SampleDict, noiseTF)
            # print("4", DMLDict)
            S1_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S1)
            result_S1 = SampleDict[DML_modules.keygen(S1_sorted)]
            # result_S1 = DML_modules.computeDML(X_train, y_train, X_test, y_test, x, S1, parentDict, topoSort)
        # print("S1:", S1)
        shapley_dict_perm[updated_variable] = result_S1 - result_S0

    return shapley_dict_perm, SampleDict

def computeShapley_Sample(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter, noiseTF):
    SampleDict = dict()
    shapley_dict = dict()

    # allCondModel trained using Train and Test dataset
    allCondModel_train = DML_modules.train_condProb(X_train, topoSort)
    allCondModel_test = DML_modules.train_condProb(X_test, topoSort)

    # Evaluated {P(Vi | pre(Vi)) } trained using Train and Test dataset
    X_list = [X_train, X_test]
    y_list = [y_train, y_test]

    model_Y_train = DML_modules.train_Y(X_train,y_train)
    model_Y_test = DML_modules.train_Y(X_test, y_test)

    allCondModel_list = [allCondModel_train, allCondModel_test]
    model_Y_list = [model_Y_train, model_Y_test]

    # prevDict = dict()
    for key in X_train.columns:
        shapley_dict[key] = 0
        # prevDict[key] = 0

    for idx in range(numiter):
        # print(idx)
        Slist = generateRandomPerm(X_train)
        shapley_dict_S, SampleDict = Shapley_update_perm_Sample(X_list, y_list, allCondModel_list, model_Y_list, x, y, Slist, topoSort, parentDict, noiseTF, SampleDict)

        for key in X_train.columns:
            shapley_dict[key] = ((idx * shapley_dict[key]) + shapley_dict_S[key]) / (idx + 1)

    return shapley_dict

def multi_run_wrapper_Sample(args):
    return computeShapley_Sample(*args)

def Run_Sample(n,seednum,numiter_per_sim,num_sim,targetidx,num_process,noiseTF):
    X, topoSort, dictName, parentDict = data_generator_1.dataGen(n, seednum)
    X_train, y_train, X_test, y_test = datagen_modules.dataSplit(X, seednum)
    x = X.iloc[[targetidx]]
    y = x[x.columns[-1]].values[0]
    x = x[x.columns[:-1]]

    myparam = (X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter_per_sim)
    numiter = numiter_per_sim * num_sim
    resultShapley = computeShapley_Sample(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter, noiseTF)

    # myparam = (X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter_per_sim)

    # pool = mp.Pool(processes=num_process)
    # results = pool.map(multi_run_wrapper_Sample, [myparam] * num_sim)
    # resultShapley = combineResult(results)
    # for idx in range(len(X_train.columns)):
    #     print(dictName[idx], ":", np.round(resultShapley[idx], 4))
    return resultShapley


if __name__ == '__main__':
    n = int(sys.argv[1])
    seednum = 123
    targetidx = 10
    num_process = 4
    
    numiter_per_sim = 5
    num_sim = 5

    casenum = int(sys.argv[2])
    noiseTF = int(sys.argv[3])

    X, topoSort, dictName, parentDict = data_generator_2.dataGen(n, seednum)
    X_train, y_train, X_test, y_test = datagen_modules.dataSplit(X, seednum)
    x = X_test.iloc[[targetidx]]
    y = y_test.iloc[[targetidx]].values[0]

    X_list = [X_train, X_test]
    y_list = [y_train, y_test]
    

    if casenum == 1:
        resultDML = Run_DML(n,seednum,numiter_per_sim,num_sim,targetidx,num_process,noiseTF)
        resultPI = Run_PI(n,seednum,numiter_per_sim,num_sim,targetidx,num_process,noiseTF)
        resultIPW = Run_IPW(n,seednum,numiter_per_sim,num_sim,targetidx,num_process,noiseTF)
        resultSample = Run_Sample(n,seednum,numiter_per_sim,num_sim,targetidx,num_process,noiseTF)
    else: 
        resultDML = Run_DML_case2(n,seednum,numiter_per_sim,num_sim,targetidx,num_process,noiseTF)
        resultPI = Run_PI_case2(n,seednum,numiter_per_sim,num_sim,targetidx,num_process,noiseTF)
        resultIPW = Run_IPW_case2(n,seednum,numiter_per_sim,num_sim,targetidx,num_process,noiseTF)
    
    print("DML", resultDML)
    print("PI", resultPI)
    print("IPW", resultIPW)
    if casenum == 1:
        print("Sample", resultSample)

    # resultDML = datagen_modules.normalizer(resultDML)

    # resultDML = Run_DML_case2(n, seednum, numiter_per_sim, num_sim, targetidx, num_process)
    # resultDML = datagen_modules.normalizer(resultDML)

    # Slist = generateRandomPerm(X_train)
    # IPWDict = {}
    # result_IPW, IPWDict  = Shapley_update_perm_IPW_case2(X_list, y_list, x, y, Slist, topoSort, IPWDict)
    # resultIPW = Run_IPW_case2(n, seednum, numiter_per_sim, num_sim, targetidx, num_process)
    # resultIPW = datagen_modules.normalizer(resultIPW)

    # PIDict = {}
    # numiter = 10
    # resultPI = computeShapley_PI_case2(X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter)
    # resultPI = Run_DML_case2(n, seednum, numiter_per_sim, num_sim, targetidx, num_process)
    # resultPI = datagen_modules.normalizer(resultPI)


    # sumval = sum(resultDML.values())
    # resultDML_normalize = resultDML.copy()
    # for key in resultDML:
    #     resultDML_normalize[key] = resultDML[key]/sumval


    # shapley_dict_perm, DMLDict = Shapley_update_perm_case2(X_list,y_list,model_Y_list,x,y,Slist,topoSort,DMLDict)

    # DMLDict = DML_modules.computeDML_S_Case2(X_list, y_list, model_Y_list, x, S, topoSort, DMLDict)

    # tmp = Run_DML(n,seednum,numiter_per_sim,num_sim,targetidx,num_process)

