import numpy as np
import pandas as pd
# import multiprocess as mp
import multiprocessing as mp 

import DML_Shapley_modules
import datagen_modules
import xgboost
import DML_modules
import xgboost_modules
import data_generator_2
import data_generator_1

def Identify_Pre_Vi(idx,topoSort):
    return topoSort[:topoSort.index(idx)]

def Extract_Pre_Vi(idx,topoSort,X):
    if len(Identify_Pre_Vi(idx,topoSort)) == 0:
        return []
    return X[Identify_Pre_Vi(idx,topoSort)]

def Derive_Ui(idx,topoSort,X):
    # Identify Vi
    Vi = X[idx]
    # Identify Pre(Vi)
    Pre_Vi = Extract_Pre_Vi(idx,topoSort,X)
    if len(Pre_Vi) == 0:
        Ui = Vi - np.mean(Vi)
        Ui = np.array(Ui).reshape((1, -1))
        Ui = Ui.T
        return(Ui)
    # Train fi := E[Vi | Pre(Vi)]
    regmode = xgboost_modules.checkMode(Vi)
    model = xgboost_modules.fit_xgboost(X=Pre_Vi, y=Vi, regtype=regmode)
    # Evaluate fi
    eval_model = xgboost_modules.predict_xgboost(model, X=Pre_Vi, y=np.asarray(Vi), regtype=regmode)
    # Derive Ui := Vi - fhat_i
    Ui = Vi - eval_model
    Ui = np.array(Ui).reshape((1,-1))
    Ui = Ui.T
    return(Ui)

def Construct_Umatrix(topoSort,X):
    # Construct U[V1,V2,...]
    UsDict = dict()
    for s in topoSort[:-1]:
        Us = Derive_Ui(s, topoSort, X)
        UsDict[s] = Us.flatten()
    UsDF = pd.DataFrame.from_dict(UsDict)
    return(UsDF)

def Train_Y_on_US(Us,y):
    # Ui_reshaped = np.array(Ui).reshape((1,-1))
    regmode = xgboost_modules.checkMode(y)
    model  = xgboost_modules.fit_xgboost(X=Us,y=y,regtype = regmode)
    return(model)

def Eval_Y_on_US(u,yvector,model):
    # Ui_reshaped = np.array(Ui).reshape((1,-1))
    regmode = xgboost_modules.checkMode(yvector)
    eval_model = xgboost_modules.predict_xgboost(model, X=u, y=yvector, regtype=regmode)
    return(eval_model)

def computeICC_S(X_list,y_list, UsDF_list, u, S,topoSort,ICCDict):

    # UsDF = Construct_Umatrix(topoSort,X)

    # Seperate
    # UsDF_train = UsDF.iloc[list(X_list[0].index)]
    # UsDF_test = UsDF.iloc[list(X_list[1].index)]

    y_train = y_list[0]
    y_test = y_list[1]
    UsDF_train = UsDF_list[0]
    UsDF_test = UsDF_list[1]

    # If S = emptyset or S is a all feature set,
    if len(S) == 0 or len(S) == len(X_list[0].columns):
        ICCDict[DML_modules.keygen(S)] = 1
        return ICCDict

    # Construct US
    # Preprocessing the S
    S_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S)
    # for s in S_sorted[:-1]:
        # Construct Us
    # UsDF_subsetS = UsDF[S_sorted]
    UsDF_subsetS_train = UsDF_train[S_sorted]
    UsDF_subsetS_test = UsDF_test[S_sorted]
    u_subset = u[S_sorted]
    model_S_train = Train_Y_on_US(UsDF_subsetS_train, y_train)
    model_S_test = Train_Y_on_US(UsDF_subsetS_test, y_test)

    yhat_S_train = Eval_Y_on_US(u_subset,y_train,model_S_train)
    yhat_S_test = Eval_Y_on_US(u_subset,y_test,model_S_test)
    result = (yhat_S_train[0] + yhat_S_test[0])/2
    ICCDict[DML_modules.keygen(S_sorted)]  = result
    return ICCDict



def Shapley_update_perm_ICC(X_list, y_list, UsDF_list, u, y, Slist, topoSort, ICCDict):
    '''
    Output:
        E[Y | do(v_{S1})] - E[Y | do(v_{S0})]
        Updated DMLDict
    '''
    # Dictionary Dict(Var) for individual permutation
    shapley_dict_perm = dict()
    # For each S generated for the individual permutation
    for idx in range(1,len(X_list[0].columns)+1):
        # pre_{\pi}(Vi)
        S0 = Slist[idx-1]
        # Vi and pre_{\pi}(Vi)
        S1 = Slist[idx]
        # Vi
        updated_variable = np.setdiff1d(S1,S0)[0]

        # If S0 = emptyset,
        if len(S0) == 0:
            # E[Y]
            result_S0 = (np.mean(y_list[0]) + np.mean(y_list[1]))/2
        else:
            # Update the DMLDict for S0
            ICCDict = computeICC_S(X_list,y_list, UsDF_list, u, S0,topoSort,ICCDict)
            # Have E[Y | do(v_{S0})]
            S0_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S0)
            result_S0 = ICCDict[DML_modules.keygen(S0_sorted)]

        # If S1 = [n], give y
        if len(S1) == len(X_list[0].columns):
            result_S1 = y
        else:
            # Update the DMLDict for S1
            ICCDict = computeICC_S(X_list,y_list, UsDF_list, u, S1,topoSort,ICCDict)
            # Have E[Y | do(v_{S1})]
            S1_sorted = datagen_modules.sort_on_reference(ref=topoSort, vec=S1)
            result_S1 = ICCDict[DML_modules.keygen(S1_sorted)]

        # Compute E[Y | do(v_{S1})] - E[Y | do(v_{S0})]
        shapley_dict_perm[updated_variable] = result_S1 - result_S0

    return shapley_dict_perm, ICCDict

def computeShapley_ICC(X_train, y_train, X_test, y_test, UsDF_train, UsDF_test, u, y, parentDict, topoSort, numiter):
    '''
    Output
        shapley_dict: Final Shapley Value
    '''
    # Initialize the DML_Dict
    ICCDict = dict()
    # Final Shapley Value
    shapley_dict = dict()

    # allCondModel trained using Train and Test dataset
    # allCondModel_train = DML_modules.train_condProb(X_train, topoSort)
    # allCondModel_test = DML_modules.train_condProb(X_test, topoSort)
    #
    # # Evaluated {P(Vi | pre(Vi)) } trained using Train and Test dataset
    # allCondProb_eval_test = DML_modules.estimate_condProb(X_test, allCondModel_train, topoSort)
    # allCondProb_eval_train = DML_modules.estimate_condProb(X_train, allCondModel_test, topoSort)
    X_list = [X_train, X_test]
    y_list = [y_train, y_test]
    UsDF_list = [UsDF_train, UsDF_test]
    # allCondProb_list = [allCondProb_eval_train, allCondProb_eval_test]

    for key in X_train.columns:
        shapley_dict[key] = 0

    # Repeat tje following:
    for idx in range(numiter):
        # Generate S
        Slist = DML_Shapley_modules.generateRandomPerm(X_train)
        # Generate an individual Shapley S and DML Dict.
        shapley_dict_S, DML_dict = Shapley_update_perm_ICC(X_list, y_list, UsDF_list, u, y, Slist, topoSort, ICCDict)

        # Moving Average Update for the Shapley.
        for key in X_train.columns:
            shapley_dict[key] = ((idx * shapley_dict[key]) + shapley_dict_S[key]) / (idx+1)

    return shapley_dict


def multi_run_wrapper_ICC(args):
    return computeShapley_ICC(*args)


def Run_ICC(n,seednum,numiter_per_sim,num_sim,targetidx,num_process):
    '''
    num_sim = number of parallelized worker
    numiter_per_sim = For each parallelized worker, the number of random permutation
    That is, the total number is num_sim *numiter_per_sim.
    '''
    X, topoSort, dictName, parentDict = data_generator_1.dataGen(n, seednum)
    # UsDF = Construct_Umatrix(topoSort,X)
    X_train, y_train, X_test, y_test = datagen_modules.dataSplit(X, seednum)
    UsDF = Construct_Umatrix(topoSort,X)
    UsDF_train = UsDF.iloc[list(X_train.index)]
    UsDF_test = UsDF.iloc[list(X_test.index)]

    x = X.iloc[[targetidx]]
    y = x[x.columns[-1]].values[0]
    x = x[x.columns[:-1]]
    u = UsDF.iloc[[targetidx]]

    myparam = (X_train, y_train, X_test, y_test, UsDF_train, UsDF_test, u, y, parentDict, topoSort, numiter_per_sim)

    # results = list()
    # for idx in range(num_sim*numiter_per_sim):
    #     results.append( multi_run_wrapper_ICC(myparam) )
    
    pool = mp.Pool(processes=num_process)
    results = pool.map(multi_run_wrapper_ICC, [myparam] * num_sim)
    resultShapley = DML_Shapley_modules.combineResult(results)
    # for idx in range(len(X_train.columns)):
    #     print(dictName[idx], ":", np.round(resultShapley[idx], 4))
    return resultShapley



'''
-- ICC Pseudocode -- 
1. For each Vi \in V, 
    A. Identify Pre(Vi)  
    B. Train fi := E[Vi | Pre(Vi)] 
    C. Evaluate fi. Call the output as fhat_i.  
    D. Derive Ui := Vi - fhat_i. 
2. Train gi := E[Y | Ui]
3. Evaluate gi.   
'''


if __name__ == '__main__':
    n = 10000
    seednum = 123
    targetidx = 10
    X, topoSort, dictName, parentDict = data_generator_1.dataGen(n, seednum)
    X_train, y_train, X_test, y_test = datagen_modules.dataSplit(X, seednum)
    x = X.iloc[[targetidx]]
    y = x[x.columns[-1]].values[0]
    x = x[x.columns[:-1]]

    X_list = [X_train, X_test]
    y_list = [y_train, y_test]

    Slist = DML_Shapley_modules.generateRandomPerm(X_train)
    ICCDict = {}
    tmpDict = {}
    # tmpDict, ICCDict = Shapley_update_perm_ICC(X_list, y_list, x, y, Slist, topoSort, ICCDict)

    num_process = 2
    num_sim = 5
    numiter_per_sim = 5
    result_ICC = Run_ICC(n, seednum, numiter_per_sim, num_sim, targetidx, num_process)

    # myparam = (X_train, y_train, X_test, y_test, x, y, parentDict, topoSort, numiter_per_sim)

