import numpy as np
import scipy as sp
import pandas as pd
import sklearn.model_selection
import datagen_modules

def dataGen(n,seednum, binnum = 2, S = [],vs = 0):
    # Fix the random seed
    np.random.seed(seednum)

    # Define variables
    X = pd.DataFrame()
    dictName = {}
    parentDict = {}

    ########################################################################
    # Generate dataset
    ########################################################################
    dictName[0] = "Sales calls"
    X[0] = np.random.uniform(0, 4, size=(n,)).round()
    if 0 in S:
        X[0] = np.repeat(vs[0].values[0],n)

    dictName[1] = "Interactions"
    pdvec = X[0] + np.random.poisson(0.2, size=(n,))
    X[1] = datagen_modules.discretizer( pdvec=pdvec, numbins= binnum )
    if 1 in S:
        X[1] = np.repeat(vs[1].values[0],n)

    dictName[2] = "Economy"
    pdvec = np.random.uniform(0, 1, size=(n,))
    X[2] = datagen_modules.discretizer(pdvec, numbins = binnum)
    if 2 in S:
        X[2] = np.repeat(vs[2].values[0],n)

    dictName[3] = "Last upgrade"
    pdvec = np.random.uniform(0, 20, size=(n,))
    X[3] = datagen_modules.discretizer(pdvec, numbins = binnum)
    if 3 in S:
        X[3] = np.repeat(vs[3].values[0],n)

    dictName[4] = "Product need"
    pdvec = X[0] * 0.1 + np.random.normal(0, 1, size=(n,))
    X[4] = datagen_modules.discretizer( pdvec=pdvec, numbins= binnum )
    if 4 in S:
        X[4] = np.repeat(vs[4].values[0],n)

    dictName[5] = "Discount"
    pdvec = ((1 - sp.special.expit(X[4])) * 0.5 + 0.5 * np.random.uniform(0, 1, size=(n,))) / 2
    X[5] = datagen_modules.discretizer( pdvec=pdvec, numbins=binnum)
    if 5 in S:
        X[5] = np.repeat(vs[5].values[0],n)

    dictName[6] = "Monthly usage"
    pdvec = sp.special.expit(X[4] * 0.3 + np.random.normal(0, 1, size=(n,)))
    X[6] = datagen_modules.discretizer( pdvec=pdvec, numbins=binnum)
    if 6 in S:
        X[6] = np.repeat(vs[6].values[0],n)

    dictName[7] = "Ad spend"
    pdvec = X[6] * np.random.uniform(0.99, 0.9, size=(n,)) + (X[3] < 1) + (
                X[3] < 2)
    X[7] = datagen_modules.discretizer( pdvec = pdvec, numbins= binnum)
    if 7 in S:
        X[7] = np.repeat(vs[7].values[0],n)

    dictName[8] = "Bugs"
    pdvec = np.array([np.random.poisson(v * 2) for v in X[6]])
    X[8] = datagen_modules.discretizer(pdvec = pdvec, numbins=binnum)
    if 8 in S:
        X[8] = np.repeat(vs[8].values[0],n)

    dictName[9] = "Did renew"
    # 4,6,
    X[9] = 7 * (
            0.9 * np.random.normal(X[4], 1, size=(n,)) \
            + 0.8 * np.random.normal(X[6], 1, size=(n,)) \
            - 0.2 * np.random.normal(X[2], 1, size=(n,)) \
            + 0.05 * np.random.normal(X[5], 1, size=(n,)) \
            - 0.6 * np.random.normal(0, 1, size=(n,)) \
            - 0.3 * (1 - np.random.normal(X[8], 1, size=(n,)) / 20) \
            + 0.2 * np.random.normal(X[0], 1, size=(n,)) \
            + 0.3 * X[1] \
            # + 0.5 / (X[3] / 4 + 0.25)
            + 0.5 * (X[3]  + 0.25)
            + X[7] * 0.6 - 0.45
    )
    X[9] = sp.special.expit(X[9])
    # X[9] = sp.stats.norm.rvs(loc=X[9], scale=1)

    # X[9] = sp.special.expit(7 * (
    #         0.9 * X[4] \
    #         + 0.8 * X[6] \
    #         + 0.2 * X[2] \
    #         + 0.05 * X[5] \
    #         + 0.05 * np.random.normal(0, 1, size=(n,)) \
    #         + 0.05 * (1 - X[8] / 20) \
    #         + 0.2 * X[0] \
    #         + 0.3 * X[1] \
    #         + 0.5 / (X[3]/ 4 + 0.25)
    #         + X[7] * 0.0 - 0.45
    # ))
    # X[9] = sp.stats.bernoulli.rvs(X[9])

    ########################################################################
    # Topo order
    ########################################################################
    def deriveTopoOrder(X):
        g = datagen_modules.Graph(len(X.columns))
        for x in range(len(X.columns)):
            parentDict[x] = []

        g.addEdge(0, 1); parentDict[1].append(0)
        g.addEdge(0, 4); parentDict[4].append(0)
        g.addEdge(4, 5); parentDict[5].append(4)
        g.addEdge(4, 6); parentDict[6].append(4)
        g.addEdge(6, 7); parentDict[7].append(6)
        g.addEdge(6, 8); parentDict[8].append(6)
        g.addEdge(0, 9); parentDict[9].append(0)
        g.addEdge(1, 9); parentDict[9].append(1)
        g.addEdge(2, 9); parentDict[9].append(2)
        g.addEdge(3, 9); parentDict[9].append(3)
        g.addEdge(4, 9); parentDict[9].append(4)
        g.addEdge(5, 9); parentDict[9].append(5)
        g.addEdge(6, 9); parentDict[9].append(6)
        g.addEdge(7, 9); parentDict[9].append(7)
        g.addEdge(8, 9); parentDict[9].append(8)
        topoSort = g.topologicalSort()

        return topoSort
    topoSort = deriveTopoOrder(X)

    ########################################################################
    # ordered set
    ########################################################################
    ordered_columns = [X.columns.to_list()[x] for x in topoSort]
    X = X[ordered_columns]


    return X, topoSort, dictName, parentDict



if __name__ == '__main__':
    print("hello, world")
    n = 10000
    seednum = 1234

    X, topoSort, dictName, parentDict = dataGen(n,seednum)
    X_train, y_train, X_test, y_test = datagen_modules.dataSplit(X,seednum)


