import numpy as np
import scipy as sp
import pandas as pd
import sklearn
import xgboost
from sklearn import preprocessing


def checkMode(vec):
    if len(pd.unique(vec)) == 2:
        return "binary"
    elif len(pd.unique(vec)) > 2 and len(pd.unique(vec)) < 10:
        return "multi"
    else:
        return "continuous"

def fit_xgboost(X, y, regtype):
    """ Train an XGBoost model with early stopping.
    regtype
    * 0: binary
    * 1: multi-class
    * 2: continuous
    """
    if regtype == "multi" or regtype == "binary":
        label = y.copy()
        y_unique = sorted(pd.unique(y))
        for idx in range(len(y_unique)):
            yvar = y_unique[idx]
            label[label == yvar] = int(idx)
    else:
        label = y.copy()

    # X_train,X_test,y_train,y_test = sklearn.model_selection.train_test_split(X, label)
    X_train, X_test, y_train, y_test = X,X,label,label
    dtrain = xgboost.DMatrix(X_train, label=y_train)
    dtest = xgboost.DMatrix(X_test, label=y_test)

    if regtype == 'binary':
        model = xgboost.train(
            { "eta": 0.001, "subsample": 0.5, "max_depth": 2, "objective": "reg:logistic"}, dtrain, num_boost_round=10,
            evals=((dtest, "test"),), early_stopping_rounds=20, verbose_eval=False
        )
        return model
    elif regtype == 'multi':
        model = xgboost.train(
            {"eta": 0.001, "subsample": 0.5, "max_depth": 2, "objective": "multi:softprob",
             "num_class": len(pd.unique(label)), "eval_metric":'mlogloss'}, dtrain,
            evals=((dtest, "test"),), early_stopping_rounds=20, verbose_eval=False, num_boost_round=10,
            evals_result={'eval_metric': 'merror'}
        )
        return model
    elif regtype == 'continuous':
        model = xgboost.train(
            {"eta": 0.001, "subsample": 0.5, "max_depth": 2, "objective": "reg:squarederror"}, dtrain, num_boost_round=10,
            evals=((dtest, "test"),), early_stopping_rounds=20, verbose_eval=False
        )
        return model

def predict_xgboost(model, X, y, regtype):
    if regtype == "multi" or regtype == "binary":
        label = y.copy()
        y_unique = sorted(pd.unique(y))
        for idx in range(len(y_unique)):
            yvar = y_unique[idx]
            label[label == yvar] = idx
    else:
        # inverse label
        label = y.copy()

    if regtype == 'binary':
        pred = model.predict(xgboost.DMatrix(X))
        pred = label * pred + (1 - label) * (1 - pred)
        return pred
    if regtype == 'multi':
        pred = model.predict(xgboost.DMatrix(X))
        pred = np.asarray([pred[idx][int(label[idx])] for idx in range(X.shape[0])])
        return pred
    if regtype == 'continuous':
        pred = model.predict(xgboost.DMatrix(X))
        return pred

