import numpy as np
from scipy import stats
import random
import pickle

from module_case3 import *

def Rank_Shapley_Values(Shapley_value):
    tmp = Shapley_value.values()
    tmp_sorted = sorted(tmp)
    tmp_rank = [tmp_sorted.index(v) for v in tmp]
    return(tmp_rank)

if __name__ == '__main__':
    ''' Experimental Setup '''
    # Set the file name
    # filename = "save_results/results_case3_v220616_1135.pkl"
    filename = "save_results/results_case3_v220616_1320_params_S20_R50.pkl"

    # number of samples
    nlist = [100,250,500,750,1000]

    # number of S sampling
    num_S_sample = 20
    num_simulation_round = 50

    # Values to be analyzed
    v1 = 1
    v2 = 1
    v3 = 1
    intv_vector = [v1, v2, v3]

    ''' Experiments Scenario '''
    # scenario = 0; scenario_description = "Nothing"  # Nothing
    # scenario = 1; scenario_description = "Converging noise" # Converging noise
    # scenario = 2; scenario_description = "REG wrong"  # REG wrong
    # scenario = 3; scenario_description = "IPW wrong" # IPW wrong

    dict_scenario = {}
    dict_scenario[0] = "Nothing"
    dict_scenario[1] = "Converging noise"
    dict_scenario[2] = "REG wrong"
    dict_scenario[3] = "IPW wrong"

    # ======= ======= ======= Experiment Begin! ======= ======= =======
    '''  For each n and scenario '''
    save_results = {}
    for scenario in [0,1,2,3]:
        save_results[scenario] = {}
        for n in nlist:
            save_results[scenario][n] = {}
            save_results[scenario][n]['n'] = -1
            save_results[scenario][n]['scenario'] = -1
            save_results[scenario][n]['scenario_description'] = -1
            save_results[scenario][n]['intv_vector'] = intv_vector
            save_results[scenario][n]['rank_corr_DML'] = []
            save_results[scenario][n]['rank_corr_Cond'] = []
            save_results[scenario][n]['REG_acc'] = []
            save_results[scenario][n]['IPW_acc'] = []
            save_results[scenario][n]['DML_acc'] = []


    for scenario in [0, 1, 2, 3]:
        for n in nlist:
            scenario_description = dict_scenario[scenario]

            ''' Iterate multiple simulation given fixed n  '''
            # For each simulation, we want to save accuracies.
            save_results[scenario][n]['n'] = n
            save_results[scenario][n]['scenario'] = scenario
            save_results[scenario][n]['scenario_description'] = scenario_description


            for sim_idx in range(num_simulation_round):
                print("Scenario:", scenario, scenario_description, "; Data",n, "; Round:", sim_idx)
                # Fix the random seed for each simulation
                seednum = random.randint(1, 100000)
                np.random.seed(seednum)

                ''' Data Generation '''
                # Generate OBS data
                Data = SCM(n, seednum=seednum)
                dim_Data = Data[0].shape[1]  # dimension of the data

                # Generate noises
                converging_noise = np.random.normal(n ** (-1 / 4), n ** (-1 / 4), n)
                noise = np.random.normal(0, 1, n)
                noises = [converging_noise, noise]

                ''' Initialize box to collect experimental results '''
                # Initialize the Shapley
                Shapley_values_REG = dict()
                Shapley_values_IPW = dict()
                Shapley_values_DML = dict()
                Shapley_values_Cond = dict()
                Shapley_True = dict()

                # Initialize the Shapley values to zero
                for idx in range(1, dim_Data + 1):
                    Shapley_True[idx] = 0
                    Shapley_values_REG[idx] = 0
                    Shapley_values_IPW[idx] = 0
                    Shapley_values_DML[idx] = 0
                    Shapley_values_Cond[idx] = 0


                ''' Experiment Iteration '''
                # For each iteration
                for experiment_idx in range(num_S_sample):
                    # print("Round:", experiment_idx+1)

                    # Generate random permutation
                    permuted = np.random.permutation([1,2,3])

                    # Generate the set S
                    Slist = []
                    for idx in range(len(permuted)+1):
                        Slist.append( permuted[:idx] )

                    # Shapley value vector
                    Shapley_values_REG_experiment_idx = dict()
                    Shapley_values_IPW_experiment_idx  = dict()
                    Shapley_values_DML_experiment_idx  = dict()
                    Shapley_values_Cond_experiment_idx  = dict()
                    Shapley_True_experiment_idx  = dict()

                    # Initialize the Shapley values to zero
                    for idx in range(1,dim_Data+1):
                        Shapley_True_experiment_idx[idx] = 0
                        Shapley_values_REG_experiment_idx[idx] = 0
                        Shapley_values_IPW_experiment_idx[idx] = 0
                        Shapley_values_DML_experiment_idx[idx] = 0
                        Shapley_values_Cond_experiment_idx[idx] = 0


                    for idx in range(1,dim_Data+1): # For all S
                        # Take the variable index to be updated
                        S0 = Slist[idx - 1] # pre_{\pi}(Vi)
                        S1 = Slist[idx] # Vi and pre_{\pi}(Vi)
                        updated_idx = np.setdiff1d(S1, S0)[0] # Vi

                        ''' For S0 '''
                        # Compute Causal, Cond
                        intv_idx= list(S0)
                        intv_idx.sort()
                        intv = [intv_vector[subidx - 1] for subidx in intv_idx]
                        intervention = ["v" + str(subidx) for subidx in intv_idx]

                        do_est_S0, true_do_S0, cond_Y_S0 = Compute_Causal(scenario, Data, noises, intv_idx, intv)
                        reg_do_S0, ipw_do_S0, DML_do_S0 = do_est_S0

                        ''' For S1 '''
                        # Compute Causal, Cond
                        intv_idx = list(S1)
                        intv_idx.sort()
                        intv = [intv_vector[subidx - 1] for subidx in intv_idx]
                        intervention = ["v" + str(subidx) for subidx in intv_idx]

                        do_est_S1, true_do_S1, cond_Y_S1 = Compute_Causal(scenario, Data, noises, intv_idx, intv)
                        reg_do_S1, ipw_do_S1, DML_do_S1 = do_est_S1

                        # Compute T(vi, pre_vi) - T(pre_vi)
                        Shapley_True_experiment_idx[updated_idx] = true_do_S1 - true_do_S0
                        Shapley_values_REG_experiment_idx[updated_idx] = reg_do_S1 - reg_do_S0
                        Shapley_values_IPW_experiment_idx[updated_idx] = ipw_do_S1 - ipw_do_S0
                        Shapley_values_DML_experiment_idx[updated_idx] = DML_do_S1 - DML_do_S0
                        Shapley_values_Cond_experiment_idx[updated_idx] = cond_Y_S1 - cond_Y_S0

                    # Update the Shapley: Shap_vi += T(vi, pre_vi) - T(pre_vi)
                    for idx in range(1,dim_Data+1): # For all S
                        Shapley_True[idx] += Shapley_True_experiment_idx[idx]
                        Shapley_values_REG[idx] += Shapley_values_REG_experiment_idx[idx]
                        Shapley_values_IPW[idx] += Shapley_values_IPW_experiment_idx[idx]
                        Shapley_values_DML[idx] += Shapley_values_DML_experiment_idx[idx]
                        Shapley_values_Cond[idx] += Shapley_values_Cond_experiment_idx[idx]

                # Take the average to get the Shapely value.
                for idx in range(1, dim_Data + 1):  # For all S
                    Shapley_True[idx] /= num_S_sample
                    Shapley_values_REG[idx] /= num_S_sample
                    Shapley_values_IPW[idx] /= num_S_sample
                    Shapley_values_DML[idx] /= num_S_sample
                    Shapley_values_Cond[idx] /= num_S_sample

                # Feature Ranking
                ranking_True = Rank_Shapley_Values(Shapley_True)
                ranking_DML = Rank_Shapley_Values(Shapley_values_DML)
                ranking_Cond = Rank_Shapley_Values(Shapley_values_Cond)

                # Rank Correlation
                rank_corr_DML = stats.spearmanr(ranking_True, ranking_DML)
                rank_corr_DML = rank_corr_DML.correlation

                rank_corr_Cond = stats.spearmanr(ranking_True, ranking_Cond)
                rank_corr_Cond = rank_corr_Cond .correlation

                # Compute the accuracy
                REG_acc = np.mean(np.abs(np.asarray(list(Shapley_True.values())) - np.asarray(list(Shapley_values_REG.values()))))
                IPW_acc = np.mean(np.abs(np.asarray(list(Shapley_True.values())) - np.asarray(list(Shapley_values_IPW.values()))))
                DML_acc = np.mean( np.abs( np.asarray(list(Shapley_True.values())) - np.asarray(list(Shapley_values_DML.values())) ) )

                save_results[scenario][n]['rank_corr_DML'].append(rank_corr_DML)
                save_results[scenario][n]['rank_corr_Cond'].append(rank_corr_Cond)
                save_results[scenario][n]['REG_acc'].append(REG_acc)
                save_results[scenario][n]['IPW_acc'].append(IPW_acc)
                save_results[scenario][n]['DML_acc'].append(DML_acc)

    pickle.dump(save_results, open(filename,'wb'))



