import pickle
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

''' Final results -- Case 1 '''
# filename = "save_results/results_case1_v220616_1300_params_S20_R50.pkl"
# fig_name = "save_results/figures_case1_v220616_1300_params_S20_R50.png"

''' Final results -- Case 2 '''
filename = "save_results/results_case2_v220616_1315_params_S20_R50.pkl"
fig_name = "save_results/figures_case2_v220616_1315_params_S20_R50.png"

''' Final results -- Case 3 '''
# filename = "save_results/results_case3_v220616_1320_params_S20_R50.pkl"
# fig_name = "save_results/figures_case3_v220616_1320_params_S20_R50.png"

# filename = "save_results/results_case2_v220616_0746.pkl"
# fig_name = "save_results/figures_case2_v220616_0746.png"

# filename = "save_results/results_case3_v220616_1135.pkl"
# fig_name = "save_results/figures_case3_v220616_1135.png"


''' Setup'''
nlist = [100,250,500,750,1000]
round_digit = 3
fixed_scenario = 0

loaded_results = pickle.load(open(filename,'rb'))


for n in nlist:
    DML_acc = np.round( np.mean(loaded_results[fixed_scenario][n]['DML_acc']), round_digit )
    REG_acc = np.round( np.mean(loaded_results[fixed_scenario][n]['REG_acc']), round_digit )
    IPW_acc = np.round( np.mean(loaded_results[fixed_scenario][n]['IPW_acc']), round_digit )

    rank_corr_DML = np.mean(loaded_results[fixed_scenario][n]['rank_corr_DML'])
    rank_corr_Cond = np.mean(loaded_results[fixed_scenario][n]['rank_corr_Cond'])

    print(DML_acc, REG_acc, IPW_acc, rank_corr_DML, rank_corr_Cond)

performance_DML = {}
performance_REG = {}
performance_IPW = {}
performance_Total = {}
for scenario in [0,1,2,3]:
    performance_DML[scenario] = []
    performance_IPW[scenario] = []
    performance_REG[scenario] = []
    for n in nlist:
        performance_DML[scenario] += loaded_results[scenario][n]['DML_acc']
        performance_IPW[scenario] += loaded_results[scenario][n]['IPW_acc']
        performance_REG[scenario] += loaded_results[scenario][n]['REG_acc']
    performance_Total[scenario] = performance_DML[scenario] + performance_IPW[scenario] + performance_REG[scenario]



Nlist_model = []
model_list = ['DML'] * len(performance_DML[scenario]) + ['IPW'] * len(performance_IPW[scenario]) + ['REG'] * len(performance_REG[scenario])

for n in nlist:
    Nlist_model += [n] * len(loaded_results[fixed_scenario][n]['DML_acc'])
Nlist = Nlist_model * 3


L1DF = dict()
L1DF['L1 error'] = performance_Total[fixed_scenario]
L1DF['N'] = Nlist
L1DF['model'] = model_list

if fixed_scenario == 0 or fixed_scenario == 1:
    sns.set(font_scale=1)
    sns.set(rc={'figure.figsize': (15, 10)})
    sns.set_style("whitegrid")

    ax_L1 = sns.boxplot(x="N", y="L1 error", hue="model", data=L1DF, palette="Set3")
    ax_L1.set(ylim=(0, 1))

    ax_L1.set_xlabel("N", fontsize=35)
    ax_L1.set_xticklabels([100, 250, 500, 750, 1000], size=25)

    ax_L1.set_ylabel("L1", fontsize=35)
    ax_L1.set_yticklabels([0, 0.2, 0.4, 0.6, 0.8, 1.0], size=25)

    # ax_L1.yaxis.grid(True)  # Hide the horizontal gridlines
    # ax_L1.xaxis.grid(True)  # Show the vertical gridlines

    if fixed_scenario == 0:
        plt.legend(loc='upper right', title='Model')
        plt.setp(ax_L1.get_legend().get_texts(), fontsize='22')  # for legend text
        plt.setp(ax_L1.get_legend().get_title(), fontsize='32')  # for legend title
    else:
        plt.legend([], [], frameon=False)

    ax_L1.figure.savefig(fig_name)

if fixed_scenario == 2 or fixed_scenario == 3:
    f, (ax, ax2) = plt.subplots(2, 1, sharex=True)

    sns.set(font_scale = 1)
    sns.set(rc={'figure.figsize': (15, 10)})
    sns.set_style("whitegrid")
    sns.boxplot(x="N", y="L1 error", hue="model", data=L1DF, palette="Set3", ax=ax)
    sns.boxplot(x="N", y="L1 error", hue="model", data=L1DF, palette="Set3", ax=ax2)
    # ax_L1.set(ylim=(0, 2))

    ax2.set_xlabel("N", fontsize=35)
    ax2.set_xticklabels(nlist, size=25)

    # zoom-in / limit the view to different portions of the data
    ax.set_ylim(1.25, 1.75)  # outliers only
    ax2.set_ylim(0, 0.3)  # most of the data

    ax2.get_legend().remove()
    ax.set_ylabel("L1", fontsize=35)
    ax.set_yticklabels([1.3, 1.4, 1.5, 1.6, 1.7], size=25)
    ax2.set_yticklabels([0, 0.05, 0.1, 0.15, 0.2, 0.25, 0.30], size=25)

    # hide the spines between ax and ax2
    ax.spines['bottom'].set_visible(False)
    ax2.spines['top'].set_visible(False)
    ax.xaxis.tick_top()
    ax.tick_params(labeltop=False)  # don't put tick labels at the top
    ax2.xaxis.tick_bottom()

    d = .015 # how big to make the diagonal lines in axes coordinates
    # arguments to pass to plot, just so we don't keep repeating them
    kwargs = dict(transform=ax.transAxes, color='k', clip_on=False)
    ax.plot((-d, +d), (-d, +d), **kwargs)        # top-left diagonal
    ax.plot((1 - d, 1 + d), (-d, +d), **kwargs)  # top-right diagonal

    kwargs.update(transform=ax2.transAxes)  # switch to the bottom axes
    ax2.plot((-d, +d), (1 - d, 1 + d), **kwargs)  # bottom-left diagonal
    ax2.plot((1 - d, 1 + d), (1 - d, 1 + d), **kwargs)  # bottom-right diagonal

    # plt.setp(ax.get_legend().get_texts(), fontsize='22')  # for legend text
    # plt.setp(ax.get_legend().get_title(), fontsize='32')  # for legend title

    ax.legend().set_visible(False)
    ax2.legend().set_visible(False)


    f.savefig(fig_name)