from sklearn.datasets import make_regression
from sklearn.ensemble import GradientBoostingRegressor
from data_gen_case3 import *

'''
######## V1 ########
'''
# E[Y | do(v1)] Estimator
def Compute_Causal_v1(scenario, Data, noises, intv):
    # Unpack the given arguments
    converging_noise, noise = noises
    v1 = intv[0]

    V, Y = Data
    V1 = np.array(V[:, 0].T)[0]
    V2 = np.array(V[:, 1].T)[0]
    V3 = np.array(V[:, 2].T)[0]

    n = V.shape[0]

    ''' Estimate E[Y | do(v1)] '''
    ######## REG for E[Y | do(v1)] ########
    # Train E[Y | V1, V2, V3 ] by regressing Y on V
    reg = GradientBoostingRegressor(random_state=0)
    if scenario == 0 or scenario == 3:
        reg.fit(V, Y)
    elif scenario == 1:
        reg.fit(V, Y + converging_noise)
    elif scenario == 2:
        reg.fit(V, noise)

    # Evaluate E[Y | v1, V2, V3]
    ## Construct v1, V2, V3
    V_v1 = np.matrix((np.repeat(v1, n), V2, V3)).T
    reg_Y_Vv1 = reg.predict(V_v1)
    reg_Y_V = reg.predict(V)

    # Train E[ E[Y | v1, V2, V3] | V1, V3] by regressing reg_Y_Vv1 to {V1,V3}
    V1_V3 = np.matrix((V1, V3)).T
    reg.fit(V1_V3, reg_Y_Vv1)

    # Evaluate E[ E[Y | v1, V2, V3] | v1, V3]
    V1_V3_v1 = np.matrix((np.repeat(v1, n), V3)).T
    reg_to_v1V3 = reg.predict(V1_V3_v1)
    reg_do_v1 = np.mean(reg_to_v1V3)

    ######## IPW for E[Y | do(v1)] ########
    # Train P(V1 | V3)
    reg.fit(V3.reshape(-1, 1), V1)

    # Evaluate P(V1 | V3)
    p_v1_v3 = reg.predict(V3.reshape(-1, 1))
    p_v1_v3[p_v1_v3 <= 0] = np.mean(p_v1_v3)  # Imputation

    # I_{v1}(V1)
    I_v1_V1 = (V1 == v1) * 1

    # I_{v1}(V1) / P(V1 | V3)
    inv_weight = I_v1_V1 / p_v1_v3
    if scenario == 1:
        inv_weight = inv_weight + converging_noise
    elif scenario == 3:
        inv_weight = noise

    # mean( inv_weight * Y)
    ipw_do_v1 = np.mean(inv_weight * Y)

    ######## DML for E[Y | do(v1)] ########
    DML_do_v1 = np.mean(inv_weight * (Y - reg_Y_V) + reg_Y_Vv1)


    return([reg_do_v1, ipw_do_v1, DML_do_v1])

# E[Y | do(v1)] True
def Compute_True_v1(intv,seednum=1):
    v1 = intv[0]
    # Generate do(v1) data
    Data_v1 = SCM(n=100000, v1=v1, seednum=seednum)
    Y_do_v1 = Data_v1[1]
    true_do_v1 = np.mean(Y_do_v1)
    return(true_do_v1)

# E[Y | v1]
def Compute_Conditional_v1(Data, intv):
    v1 = intv[0]

    V, Y = Data
    V1 = np.array(V[:, 0].T)[0]
    V2 = np.array(V[:, 1].T)[0]
    V3 = np.array(V[:, 2].T)[0]

    n = V.shape[0]

    ''' Conditional Effect -- Estimate E[Y | v1] '''
    ######## REG for E[Y | v1] ########
    reg = GradientBoostingRegressor(random_state=0)
    # Train E[Y | V1]
    reg.fit(V1.reshape(-1, 1), Y)
    # Eval E[Y | v1]
    V_v1 = np.matrix((np.repeat(v1, n))).T
    reg_Y_v1 = reg.predict(V_v1)
    cond_Y_v1 = np.mean(reg_Y_v1)
    return(cond_Y_v1)

'''
######## V3 ########
'''
def Compute_Causal_v3(scenario, Data, noises, intv):
    # Unpack the given arguments
    converging_noise, noise = noises
    v3 = intv[0]

    V, Y = Data
    V1 = np.array(V[:, 0].T)[0]
    V2 = np.array(V[:, 1].T)[0]
    V3 = np.array(V[:, 2].T)[0]

    n = V.shape[0]

    ######## REG for E[Y | do(v3)] ########
    # Train E[Y | V1, V2, V3 ] by regressing Y on V
    reg = GradientBoostingRegressor(random_state=0)
    if scenario == 0 or scenario == 3:
        reg.fit(V, Y)
    elif scenario == 1:
        reg.fit(V, Y + converging_noise)
    elif scenario == 2:
        reg.fit(V, noise)

    # Evaluate E[Y | v1, V2, V3]
    ## Construct V1, V2, v3
    V_v3 = np.matrix((V1, V2, np.repeat(v3, n))).T
    reg_Y_Vv3 = reg.predict(V_v3)
    reg_Y_V = reg.predict(V)

    # Train E[ E[Y | V1, V2, v3] | V1, V3] by regressing reg_Y_Vv3 to {V1,V3}
    V1_V3 = np.matrix((V1, V3)).T
    reg.fit(V1_V3, reg_Y_Vv3)

    # Evaluate E[ E[Y | v1, V2, V3] | V1, v3]
    V1_V3_v3 = np.matrix((V1, np.repeat(v3, n))).T
    reg_to_V1v3 = reg.predict(V1_V3_v3)
    reg_do_v3 = np.mean(reg_to_V1v3)

    ######## IPW for E[Y | do(v3)] ########
    # Train P(V3 | V1)
    reg.fit(V1.reshape(-1, 1), V3)

    # Evaluate P(V3 | V1)
    p_v3_v1 = reg.predict(V1.reshape(-1, 1))
    p_v3_v1[p_v3_v1 <= 0] = np.mean(p_v3_v1)  # Imputation

    # I_{v3}(V3)
    I_v3_V3 = (V3 == v3) * 1

    # I_{v3}(V3) / P(V3 | V1)
    inv_weight = I_v3_V3 / p_v3_v1
    if scenario == 1:
        inv_weight = inv_weight + converging_noise
    elif scenario == 3:
        inv_weight = noise

    # mean( inv_weight * Y)
    ipw_do_v3 = np.mean(inv_weight * Y)

    ######## DML for E[Y | do(v3)] ########
    DML_do_v3 = np.mean(inv_weight * (Y - reg_Y_V) + reg_Y_Vv3)
    return ([reg_do_v3, ipw_do_v3, DML_do_v3])

def Compute_True_v3(intv,seednum=1):
    v3 = intv[0]
    # Generate do(v3) data
    Data_v3 = SCM(n=100000, v3=v3, seednum=seednum)
    Y_do_v3 = Data_v3[1]
    true_do_v3 = np.mean(Y_do_v3)
    return(true_do_v3)

def Compute_Conditional_v3(Data,intv):
    # Unpack the given arguments
    v3 = intv[0]

    V, Y = Data
    V1 = np.array(V[:, 0].T)[0]
    V2 = np.array(V[:, 1].T)[0]
    V3 = np.array(V[:, 2].T)[0]

    n = V.shape[0]

    ''' Conditional Effect -- Estimate E[Y | v3] '''
    ######## REG for E[Y | v3] ########
    reg = GradientBoostingRegressor(random_state=0)
    # Train E[Y | V3]
    reg.fit(V3.reshape(-1, 1), Y)
    # Eval E[Y | v3]
    V_v3 = np.matrix((np.repeat(v3, n))).T
    reg_Y_v3 = reg.predict(V_v3)
    cond_Y_v3 = np.mean(reg_Y_v3)
    return(cond_Y_v3)

'''
######## V2 ########
'''
def Compute_Causal_v2(scenario, Data, noises, intv):
    # Unpack the given arguments
    converging_noise, noise = noises
    v2 = intv[0]

    V, Y = Data
    V1 = np.array(V[:, 0].T)[0]
    V2 = np.array(V[:, 1].T)[0]
    V3 = np.array(V[:, 2].T)[0]

    n = V.shape[0]

    ######## REG for E[Y | do(v2)] ########
    # Train E[Y | V1, V2, V3 ] by regressing Y on V
    reg = GradientBoostingRegressor(random_state=0)
    if scenario == 0 or scenario == 3:
        reg.fit(V, Y)
    elif scenario == 1:
        reg.fit(V, Y + converging_noise)
    elif scenario == 2:
        reg.fit(V, noise)

    # Evaluate E[Y | V1, v2, V3]
    ## Construct V1, v2, V3
    V_v2 = np.matrix((V1, np.repeat(v2, n), V3)).T
    reg_Y_Vv2 = reg.predict(V_v2)
    reg_Y_V = reg.predict(V)

    ## Evaluate E[Y | do(v2)) = mean( E[Y | V1, v2, V3] )
    reg_do_v2 = np.mean(reg_Y_Vv2)

    ######## IPW for E[Y | do(v2)] ########
    # Train P(V2 | V1,V3)
    reg.fit(np.matrix((V1, V3)).T, V2)

    # Evaluate P(V2 | V1, V3)
    p_v2_v1v3 = reg.predict(np.matrix((V1, V3)).T)
    p_v2_v1v3[p_v2_v1v3 <= 0] = np.mean(p_v2_v1v3)  # Imputation

    # I_{v2}(V2)
    I_v2_V2 = (V2 == v2) * 1

    # I_{v2}(V2) / P(V2 | V1, V3)
    inv_weight = I_v2_V2 / p_v2_v1v3
    if scenario == 1:
        inv_weight = inv_weight + converging_noise
    elif scenario == 3:
        inv_weight = noise

    # mean( inv_weight * Y)
    ipw_do_v2 = np.mean(inv_weight * Y)

    ######## DML for E[Y | do(v2)] ########
    DML_do_v2 = np.mean(inv_weight * (Y - reg_Y_V) + reg_Y_Vv2)

    return([reg_do_v2, ipw_do_v2, DML_do_v2])

def Compute_True_v2(intv,seednum=1):
    v2 = intv[0]
    # Generate do(v2) data
    Data_v2 = SCM(n=100000, v2=v2, seednum=seednum)
    Y_do_v2 = Data_v2[1]
    true_do_v2 = np.mean(Y_do_v2)
    return(true_do_v2)

def Compute_Conditional_v2(Data,intv):
    # Unpack the given arguments
    v2 = intv[0]

    V, Y = Data
    V1 = np.array(V[:, 0].T)[0]
    V2 = np.array(V[:, 1].T)[0]
    V3 = np.array(V[:, 2].T)[0]

    n = V.shape[0]

    ''' Conditional Effect -- Estimate E[Y | v2] '''
    ######## REG for E[Y | v2] ########
    reg = GradientBoostingRegressor(random_state=0)
    # Train E[Y | V2]
    reg.fit(V2.reshape(-1, 1), Y)
    # Eval E[Y | v2]
    V_v2 = np.matrix((np.repeat(v2, n))).T
    reg_Y_v2 = reg.predict(V_v2)
    cond_Y_v2 = np.mean(reg_Y_v2)
    return(cond_Y_v2)


'''
######## V1, V2 ########
'''
def Compute_Causal_v1v2(scenario, Data, noises, intv):
    # Unpack the given arguments
    converging_noise, noise = noises
    v1, v2 = intv

    V, Y = Data
    V1 = np.array(V[:, 0].T)[0]
    V2 = np.array(V[:, 1].T)[0]
    V3 = np.array(V[:, 2].T)[0]

    n = V.shape[0]

    ######## REG for E[Y | do(v1,v2)] ########
    # Train E[Y | V1, V2, V3 ] by regressing Y on V
    reg = GradientBoostingRegressor(random_state=0)
    if scenario == 0 or scenario == 3:
        reg.fit(V, Y)
    elif scenario == 1:
        reg.fit(V, Y + converging_noise)
    elif scenario == 2:
        reg.fit(V, noise)

    # Evaluate E[Y | v1, v2, V3]
    ## Construct v1, v2, V3
    V_v1v2 = np.matrix((np.repeat(v1, n), np.repeat(v2, n), V3)).T
    reg_Y_Vv1v2 = reg.predict(V_v1v2)
    reg_Y_V = reg.predict(V)

    ## Evaluate E[Y | do(v1,v2)) = mean( E[Y | v1, v2, V3] )
    reg_do_v1v2 = np.mean(reg_Y_Vv1v2)

    ######## IPW for E[Y | do(v1,v2)] ########
    # Train P(V1,V2 | V3) = P(V2 | V1, V3) * P(V1 | V3)
    ## Train P(V2 | V1, V3)
    reg.fit(np.matrix((V1, V3)).T, V2)
    ## Evaluate P(V2 | V1, V3)
    p_v2_v1v3 = reg.predict(np.matrix((V1, V3)).T)
    p_v2_v1v3[p_v2_v1v3 <= 0] = np.mean(p_v2_v1v3)  # Imputation

    ## Train P(V1 | V3)
    reg.fit(V3.reshape(-1, 1), V1)
    ## Evaluate P(V1 | V3)
    p_v1_v3 = reg.predict(V3.reshape(-1, 1))
    p_v1_v3[p_v1_v3 <= 0] = np.mean(p_v1_v3)  # Imputation

    # Evaluate P(V1,V2 | V3) = P(V2 | V1, V3) * P(V1 | V3)
    p_v1v2_v3 = p_v2_v1v3 * p_v1_v3

    # I_{v1v2}(V1V2)
    I_v1v2_V1V2 = ((V1 == v1) * 1) * ((V2 == v2) * 1)

    # I_{v1v2}(V1V2) / P(V1,V2 | V3)
    inv_weight = I_v1v2_V1V2 / p_v1v2_v3
    if scenario == 1:
        inv_weight = inv_weight + converging_noise
    elif scenario == 3:
        inv_weight = noise

    # mean( inv_weight * Y)
    ipw_do_v1v2 = np.mean(inv_weight * Y)

    ######## DML for E[Y | do(v1,v2)] ########
    DML_do_v1v2 = np.mean(inv_weight * (Y - reg_Y_V) + reg_Y_Vv1v2)

    return([reg_do_v1v2, ipw_do_v1v2, DML_do_v1v2])

def Compute_True_v1v2(intv,seednum=1):
    v1,v2 = intv
    # Generate do(v1, v2) data
    Data_v1v2 = SCM(n=100000, v1=v1, v2=v2, seednum=seednum)
    Y_do_v1v2 = Data_v1v2[1]
    true_do_v1v2 = np.mean(Y_do_v1v2)
    return(true_do_v1v2)

def Compute_Conditional_v1v2(Data, intv):
    # Unpack the given arguments
    v1, v2 = intv

    V, Y = Data
    V1 = np.array(V[:, 0].T)[0]
    V2 = np.array(V[:, 1].T)[0]
    V3 = np.array(V[:, 2].T)[0]

    n = V.shape[0]
    ''' Conditional Effect -- Estimate E[Y | v1, v2] '''
    ######## REG for E[Y | v2] ########
    reg = GradientBoostingRegressor(random_state=0)
    # Train E[Y | V1, V2]
    reg.fit(np.matrix((V1, V2)).T, Y)
    # Eval E[Y | v2]
    V_v1v2 = np.matrix((np.repeat(v1, n), np.repeat(v2, n))).T
    reg_Y_v1v2 = reg.predict(V_v1v2)
    cond_Y_v1v2 = np.mean(reg_Y_v1v2)
    return(cond_Y_v1v2)


'''
######## V1, V3 ########
'''
def Compute_Causal_v1v3(scenario, Data, noises, intv):
    # Unpack the given arguments
    converging_noise, noise = noises
    v1, v3 = intv

    V, Y = Data
    V1 = np.array(V[:, 0].T)[0]
    V2 = np.array(V[:, 1].T)[0]
    V3 = np.array(V[:, 2].T)[0]

    n = V.shape[0]

    ######## REG for E[Y | do(v1,v3)] ########
    # Train E[Y | V1, V2, V3 ] by regressing Y on V
    reg = GradientBoostingRegressor(random_state=0)
    if scenario == 0 or scenario == 3:
        reg.fit(V, Y)
    elif scenario == 1:
        reg.fit(V, Y + converging_noise)
    elif scenario == 2:
        reg.fit(V, noise)

    # Evaluate E[Y | v1, V2, v3]
    ## Construct v1, V2, v3
    V_v1v3 = np.matrix((np.repeat(v1, n), V2, np.repeat(v3, n))).T
    reg_Y_Vv1v3 = reg.predict(V_v1v3)
    reg_Y_V = reg.predict(V)

    # Evaluate E[Y | do(v1,v3)) = E[ E[Y | v1, V2, v3] | v1, v3]
    V_V1V3 = np.matrix((V1, V3)).T
    ## Train E[ E[Y | v1, V2, v3] | V1, V3]
    reg.fit(V_V1V3, reg_Y_Vv1v3)

    ## Evaluate E[Y | do(v1,v3)] = E[ E[Y | v1, V2, v3] | v1, v3]
    ### Construct  v1,v3
    V_v1v3 = np.matrix((np.repeat(v1, n), np.repeat(v3, n))).T
    reg_do_v1v3 = np.mean(reg.predict(V_v1v3))

    ######## IPW for E[Y | do(v1,v3)] ########
    # Train P(V1,V3) = P(V3) P(V1 | V3)
    ## Train P(V1 | V3)
    reg.fit(V3.reshape(-1, 1), V1)
    ## Evaluate P(V1 | V3)
    p_v1_v3 = reg.predict(V3.reshape(-1, 1))
    p_v1_v3[p_v1_v3 <= 0] = np.mean(p_v1_v3)  # Imputation

    ## Train P(V3) & Evaluate P(V3)
    p_v3 = V3 * np.mean(V3) + (1 - V3) * (1 - np.mean(V3))

    ## Train P(V1,V3) = P(V3) P(V1 | V3)
    p_v1v3 = p_v1_v3 * p_v3

    # I_{v1v3}(V1V3)
    I_v1v3_V1V3 = ((V1 == v1) * 1) * ((V3 == v3) * 1)

    # I_v1v3_V1V3 / P(V1,V3 )
    inv_weight = I_v1v3_V1V3 / p_v1v3
    if scenario == 1:
        inv_weight = inv_weight + converging_noise
    elif scenario == 3:
        inv_weight = noise

    # mean( inv_weight * Y)
    ipw_do_v1v3 = np.mean(inv_weight * Y)

    ######## DML for E[Y | do(v1,v2)] ########
    DML_do_v1v3 = np.mean(inv_weight * (Y - reg_Y_V) + reg_do_v1v3)
    return([reg_do_v1v3, ipw_do_v1v3, DML_do_v1v3])

def Compute_True_v1v3(intv,seednum=1):
    v1,v3 = intv

    # Generate do(v1, v3) data
    Data_v1v3 = SCM(n=100000, v1=v1, v3=v3, seednum=seednum)
    Y_do_v1v3 = Data_v1v3[1]
    true_do_v1v3 = np.mean(Y_do_v1v3)
    return(true_do_v1v3)

def Compute_Conditional_v1v3(Data, intv):
    # Unpack the given arguments
    v1, v3 = intv

    V, Y = Data
    V1 = np.array(V[:, 0].T)[0]
    V2 = np.array(V[:, 1].T)[0]
    V3 = np.array(V[:, 2].T)[0]

    n = V.shape[0]

    ''' Conditional Effect -- Estimate E[Y | v1, v3] '''
    ######## REG for E[Y | v3] ########
    reg = GradientBoostingRegressor(random_state=0)
    # Train E[Y | V1, V3]
    reg.fit(np.matrix((V1, V3)).T, Y)
    # Eval E[Y | v3]
    V_v1v3 = np.matrix((np.repeat(v1, n), np.repeat(v3, n))).T
    reg_Y_v1v3 = reg.predict(V_v1v3)
    cond_Y_v1v3 = np.mean(reg_Y_v1v3)
    return(cond_Y_v1v3)

'''
######## V2, V3 ########
'''
def Compute_Causal_v2v3(scenario, Data, noises, intv):
    # Unpack the given arguments
    converging_noise, noise = noises
    v2, v3 = intv

    V, Y = Data
    V1 = np.array(V[:, 0].T)[0]
    V2 = np.array(V[:, 1].T)[0]
    V3 = np.array(V[:, 2].T)[0]

    n = V.shape[0]

    ######## REG for E[Y | do(v2,v3)] ########
    # Train E[Y | V1, V2, V3 ] by regressing Y on V
    reg = GradientBoostingRegressor(random_state=0)
    if scenario == 0 or scenario == 3:
        reg.fit(V, Y)
    elif scenario == 1:
        reg.fit(V, Y + converging_noise)
    elif scenario == 2:
        reg.fit(V, noise)

    # Evaluate E[Y | V1, v2, v3]
    ## Construct V1, v2, v3
    V_v2v3 = np.matrix((V1, np.repeat(v2, n), np.repeat(v3, n))).T
    reg_Y_Vv2v3 = reg.predict(V_v2v3)
    reg_Y_V = reg.predict(V)

    ## Evaluate E[Y | do(v2,v3)) = mean( E[Y | V1, v2, v3] )
    reg_do_v2v3 = np.mean(reg_Y_Vv2v3)

    ######## IPW for E[Y | do(v2,v3)] ########
    # Train P(V2,V3 | V1) = P(V2 | V1, V3) * P(V3 | V1)
    ## Train P(V2 | V1, V3)
    reg.fit(np.matrix((V1, V3)).T, V2)
    ## Evaluate P(V2 | V1, V3)
    p_v2_v1v3 = reg.predict(np.matrix((V1, V3)).T)
    p_v2_v1v3[p_v2_v1v3 <= 0] = np.mean(p_v2_v1v3)  # Imputation

    ## Train P(V3 | V1)
    reg.fit(V1.reshape(-1, 1), V3)
    ## Evaluate P(V1 | V3)
    p_v3_v1 = reg.predict(V1.reshape(-1, 1))
    p_v3_v1[p_v3_v1 <= 0] = np.mean(p_v3_v1)  # Imputation

    # Evaluate P(V2,V3 | V1) = P(V2 | V1, V3) * P(V3 | V1)
    p_v2v3_v1 = p_v2_v1v3 * p_v3_v1

    # I_{v2v3}(V2V3)
    I_v2v3_V2V3 = ((V2 == v2) * 1) * ((V3 == v3) * 1)

    # I_v2v3_V2V3 / P(V2,V3 | V1)
    inv_weight = I_v2v3_V2V3 / p_v2v3_v1
    if scenario == 1:
        inv_weight = inv_weight + converging_noise
    elif scenario == 3:
        inv_weight = noise

    # mean( inv_weight * Y)
    ipw_do_v2v3 = np.mean(inv_weight * Y)

    ######## DML for E[Y | do(v1,v2)] ########
    DML_do_v2v3 = np.mean(inv_weight * (Y - reg_Y_V) + reg_Y_Vv2v3)
    return([reg_do_v2v3, ipw_do_v2v3, DML_do_v2v3])

def Compute_True_v2v3(intv, seednum=1):
    v2, v3 = intv

    # Generate do(v2, v3) data
    Data_v2v3 = SCM(n=100000, v2=v2, v3=v3, seednum=seednum)
    Y_do_v2v3 = Data_v2v3[1]
    true_do_v2v3 = np.mean(Y_do_v2v3)
    return (true_do_v2v3)

def Compute_Conditional_v2v3(Data,intv):
    # Unpack the given arguments
    v2, v3 = intv

    V, Y = Data
    V1 = np.array(V[:, 0].T)[0]
    V2 = np.array(V[:, 1].T)[0]
    V3 = np.array(V[:, 2].T)[0]

    n = V.shape[0]

    ''' Conditional Effect -- Estimate E[Y | v2, v3] '''
    ######## REG for E[Y | v2] ########
    reg = GradientBoostingRegressor(random_state=0)
    # Train E[Y | V1, V2]
    reg.fit(np.matrix((V2, V3)).T, Y)
    # Eval E[Y | v2]
    V_v2v3 = np.matrix((np.repeat(v2, n), np.repeat(v3, n))).T
    reg_Y_v2v3 = reg.predict(V_v2v3)
    cond_Y_v2v3 = np.mean(reg_Y_v2v3)
    return(cond_Y_v2v3)



'''
######## V1, V2, V3 ########
'''
def Compute_Causal_v1v2v3(scenario, Data, noises, intv):
    # Unpack the given arguments
    converging_noise, noise = noises
    v1, v2, v3 = intv

    V, Y = Data
    V1 = np.array(V[:, 0].T)[0]
    V2 = np.array(V[:, 1].T)[0]
    V3 = np.array(V[:, 2].T)[0]

    n = V.shape[0]

    ######## REG for E[Y | do(v1,v2,v3)] ########
    # Train E[Y | V1, V2, V3 ] by regressing Y on V
    reg = GradientBoostingRegressor(random_state=0)
    if scenario == 0 or scenario == 3:
        reg.fit(V, Y)
    elif scenario == 1:
        reg.fit(V, Y + converging_noise)
    elif scenario == 2:
        reg.fit(V, noise)

    # Evaluate E[Y | v1, v2, v3]
    ## Construct v1, v2, v3
    V_v1v2v3 = np.matrix((np.repeat(v1, n), np.repeat(v2, n), np.repeat(v3, n))).T
    reg_Y_Vv1v2v3 = reg.predict(V_v1v2v3)
    reg_Y_V = reg.predict(V)

    ## Evaluate E[Y | do(v1,v2,v3)) = mean( E[Y | v1, v2, v3] )
    reg_do_v1v2v3 = np.mean(reg_Y_Vv1v2v3)

    ######## IPW for E[Y | do(v1,v2,v3)] ########
    # Train P(V1,V2,V3) = P(V2 | V1, V3) * P(V1 | V3) * P(V3)
    ## Train P(V2 | V1, V3)
    reg.fit(np.matrix((V1, V3)).T, V2)
    ## Evaluate P(V2 | V1, V3)
    p_v2_v1v3 = reg.predict(np.matrix((V1, V3)).T)
    p_v2_v1v3[p_v2_v1v3 <= 0] = np.mean(p_v2_v1v3)  # Imputation

    ## Train P(V1 | V3)
    reg.fit(V3.reshape(-1, 1), V1)
    ## Evaluate P(V1 | V3)
    p_v1_v3 = reg.predict(V3.reshape(-1, 1))
    p_v1_v3[p_v1_v3 <= 0] = np.mean(p_v1_v3)  # Imputation

    ## Train and Evaluate P(V3)
    p_v3 = V3 * np.mean(V3) + (1 - V3) * (1 - np.mean(V3))

    # Evaluate P(V1,V2,V3) = P(V2 | V1, V3) * P(V1 | V3) * P(V3)
    p_v1v2v3 = p_v2_v1v3 * p_v1_v3 * p_v3

    # I_{v1v2v3}(V1V2V3)
    I_v1v2v3_V1V2V3 = ((V1 == v1) * 1) * ((V2 == v2) * 1) * ((V3 == v3) * 1)

    # I_v1v2v3_V1V2V3 / P(V1,V2,V3)
    inv_weight = I_v1v2v3_V1V2V3 / p_v1v2v3
    if scenario == 1:
        inv_weight = inv_weight + converging_noise
    elif scenario == 3:
        inv_weight = noise

    # mean( inv_weight * Y)
    ipw_do_v1v2v3 = np.mean(inv_weight * Y)

    ######## DML for E[Y | do(v1,v2)] ########
    DML_do_v1v2v3 = np.mean(inv_weight * (Y - reg_Y_V) + reg_Y_Vv1v2v3)

    return([reg_do_v1v2v3, ipw_do_v1v2v3, DML_do_v1v2v3])

def Compute_True_v1v2v3(intv,seednum=1):
    v1,v2,v3 = intv
    # Generate do(v1, v2, v3) data
    Data_v1v2v3 = SCM(n=100000, v1=v1, v2=v2, v3=v3, seednum=seednum)
    Y_do_v1v2v3 = Data_v1v2v3[1]
    true_do_v1v2v3 = np.mean(Y_do_v1v2v3)
    return(true_do_v1v2v3)

def Compute_Conditional_v1v2v3(Data, intv):
    # Unpack the given arguments
    v1, v2, v3 = intv

    V, Y = Data
    V1 = np.array(V[:, 0].T)[0]
    V2 = np.array(V[:, 1].T)[0]
    V3 = np.array(V[:, 2].T)[0]

    n = V.shape[0]

    ''' Conditional Effect -- Estimate E[Y | v1, v2, v3] '''
    ######## REG for E[Y | v2] ########
    reg = GradientBoostingRegressor(random_state=0)
    # Train E[Y | V1, V2]
    reg.fit(np.matrix((V1, V2, V3)).T, Y)
    # Eval E[Y | v2]
    V_v1v2v3 = np.matrix((np.repeat(v1, n), np.repeat(v2, n), np.repeat(v3, n))).T
    reg_Y_v1v2v3_cond = reg.predict(V_v1v2v3)
    cond_Y_v1v2v3 = np.mean(reg_Y_v1v2v3_cond)
    return(cond_Y_v1v2v3)



def Compute_Causal(scenario, Data, noises, intv_idx, intv):
    if intv_idx == [1]:
        reg_do, ipw_do, DML_do = Compute_Causal_v1(scenario, Data, noises, intv)
        true_do = Compute_True_v1(intv, seednum=1)
        cond_Y = Compute_Conditional_v1(Data, intv)
    elif intv_idx == [3]:
        reg_do, ipw_do, DML_do = Compute_Causal_v3(scenario, Data, noises, intv)
        true_do = Compute_True_v3(intv, seednum=1)
        cond_Y = Compute_Conditional_v3(Data, intv)
    elif intv_idx == [2]:
        reg_do, ipw_do, DML_do = Compute_Causal_v2(scenario, Data, noises, intv)
        true_do = Compute_True_v2(intv, seednum=1)
        cond_Y = Compute_Conditional_v2(Data, intv)
    elif intv_idx == [1,2]:
        reg_do, ipw_do, DML_do = Compute_Causal_v1v2(scenario, Data, noises, intv)
        true_do = Compute_True_v1v2(intv, seednum=1)
        cond_Y = Compute_Conditional_v1v2(Data, intv)
    elif intv_idx == [1,3]:
        reg_do, ipw_do, DML_do = Compute_Causal_v1v3(scenario, Data, noises, intv)
        true_do = Compute_True_v1v3(intv, seednum=1)
        cond_Y = Compute_Conditional_v1v3(Data, intv)
    elif intv_idx == [2,3]:
        reg_do, ipw_do, DML_do = Compute_Causal_v2v3(scenario, Data, noises, intv)
        true_do = Compute_True_v2v3(intv, seednum=1)
        cond_Y = Compute_Conditional_v2v3(Data, intv)
    elif intv_idx == [1,2,3]:
        reg_do, ipw_do, DML_do = Compute_Causal_v1v2v3(scenario, Data, noises, intv)
        true_do = Compute_True_v1v2v3(intv, seednum=1)
        cond_Y = Compute_Conditional_v1v2v3(Data, intv)
    elif intv_idx == []:
        reg_do = ipw_do = DML_do = np.mean(Data[1])
        true_do = np.mean(Data[1])
        cond_Y = np.mean(Data[1])

    do_est = [reg_do, ipw_do, DML_do]

    return ([do_est, true_do, cond_Y])


if __name__ == '__main__':
    # seednum
    seednum = 1
    np.random.seed(seednum)

    # Variables
    n = 50000

    # Values
    v1 = 1
    v2 = 0
    v3 = 0
    intv_vector = [v1,v2,v3]

    # Noise
    converging_noise = np.random.normal(n**(-1/4), n**(-1/4), n)
    noise = np.random.normal(0, 1, n)
    noises = [converging_noise, noise]

    # Experiments Scenario
    scenario = 0; scenario_description = "Nothing" # Nothing
    # scenario = 1; scenario_description = "Converging noise" # Converging noise
    # scenario = 2; scenario_description = "REG wrong" # REG wrong
    # scenario = 3; scenario_description = "IPW wrong" # IPW wrong

    # Intervention
    # intervention = "v1"; intv = [v1]
    # intervention = "v3"; intv = [v3]
    # intervention = "v2"; intv = [v2]
    # intervention = "v1v2"; intv = [v1,v2]
    # intervention = "v1v3"; intv = [v1,v3]
    # intervention = "v2v3"; intv = [v2,v3]
    # intervention = "v1v2v3"; intv = [v1,v2,v3]

    intv_idx = [1,2]
    intv = [intv_vector[idx-1] for idx in intv_idx]
    intervention = ["v"+str(idx) for idx in intv_idx]


    # Generate OBS data
    Data = SCM(n,seednum=seednum)

    do_est, true_do, cond_Y = Compute_Causal(scenario,Data,noises,intv_idx,intv)
    reg_do, ipw_do, DML_do = do_est

    # Results
    print("Intervened on:", intervention)
    print("Scenario:",scenario, scenario_description)
    print("Data:",n)
    print("True:",true_do)
    print("Reg:", reg_do)
    print("IPW:", ipw_do)
    print("DML:", DML_do)
    print("Conditional:", cond_Y)

